from typing import Any, Dict, List, Optional

from enum import Enum

from pydantic import Field

from giskard.core.core import TestResultStatusEnum
from giskard.core.validation import ConfiguredBaseModel
from giskard.utils.artifacts import serialize_parameter


class TestInputDTO(ConfiguredBaseModel):
    name: str
    value: str
    type: str
    params: List["TestInputDTO"] = Field(default_factory=list)
    is_alias: bool = False
    is_default_value: bool = False

    @classmethod
    def from_inputs_dict(cls, inputs: Dict[str, Any]) -> List["TestInputDTO"]:
        return [
            cls(name=name, value=str(serialize_parameter(value)), type=type(value).__qualname__)
            for name, value in inputs.items()
        ]


class SuiteTestDTO(ConfiguredBaseModel):
    id: Optional[int]
    testUuid: str
    functionInputs: Dict[str, TestInputDTO]
    displayName: Optional[str] = None


class TestSuiteDTO(ConfiguredBaseModel):
    name: Optional[str]
    project_key: str
    tests: List[SuiteTestDTO]
    function_inputs: List[TestInputDTO]


class TestSuiteExecutionResult(str, Enum):
    IN_PROGRESS = "IN_PROGRESS"
    CANCELLED = "CANCELLED"
    PASSED = "PASSED"
    FAILED = "FAILED"
    ERROR = "ERROR"


class MLWorkerWSTestMessageType(str, Enum):
    ERROR = "ERROR"
    INFO = "INFO"


class TestResultMessageDTO(ConfiguredBaseModel):
    type: MLWorkerWSTestMessageType
    text: str


class SaveSuiteTestExecutionDetailsDTO(ConfiguredBaseModel):
    inputs: Dict[str, List[str]]
    outputs: List[str]
    results: List[TestResultStatusEnum]
    metadata: Dict[str, List[str]]


class SaveSuiteTestExecutionDTO(ConfiguredBaseModel):
    suiteTest: SuiteTestDTO
    testUuid: str
    displayName: str
    inputs: Dict[str, str]
    arguments: Dict[str, TestInputDTO]
    messages: List[TestResultMessageDTO]
    status: TestResultStatusEnum
    metric: Optional[float]
    metricName: str
    failedIndexes: Dict[str, List[int]]
    details: Optional[SaveSuiteTestExecutionDetailsDTO]


class SaveSuiteExecutionDTO(ConfiguredBaseModel):
    suiteId: Optional[int]
    label: str
    inputs: List[TestInputDTO]
    result: TestSuiteExecutionResult
    message: str
    results: List[SaveSuiteTestExecutionDTO]
    executionDate: str
    completionDate: str


class ServerInfo(ConfiguredBaseModel):
    instanceId: Optional[str] = None
    serverVersion: Optional[str] = None
    instanceLicenseId: Optional[str] = None
    user: Optional[str] = None


class TestInfo(ConfiguredBaseModel):
    testUuid: str
    functionInputs: Dict[str, Any]
    # The following inputs are never used and are here only for
    # coherence with backend
    id: int
    displayName: Optional[str]
    test: Any


class SuiteInfo(ConfiguredBaseModel):
    name: str
    tests: List[TestInfo]
    # The following inputs are never used and are here only for
    # coherence with backend
    id: int
    projectKey: str


class ModelMetaInfo(ConfiguredBaseModel):
    id: str
    name: str
    modelType: str
    featureNames: List[str]
    threshold: Optional[float] = None
    description: Optional[str] = None
    classificationLabels: Optional[List[str]] = None
    classificationLabelsDtype: Optional[str] = None
    languageVersion: str
    language: str
    createdDate: str
    size: int
    projectId: int


class DatasetMetaInfo(ConfiguredBaseModel):
    target: Optional[str] = None
    columnTypes: Dict[str, str]
    columnDtypes: Dict[str, str]
    numberOfRows: int
    categoryFeatures: Dict[str, List[str]]
    name: Optional[str] = None
    originalSizeBytes: int
    compressedSizeBytes: int
    createdDate: str
    id: str



from typing import Optional

from ...llm.client import ChatMessage, LLMClient, get_default_client
from ...llm.errors import LLMGenerationError
from ..base import AgentAnswer
from ..question_generators.utils import parse_json_output
from .base import Metric

CORRECTNESS_EVALUATION_SYSTEM_PROMPT = """Your role is to test AI agents. Your task consists in assessing whether a agent output correctly answers a question. 
You are provided with the ground truth answer to the question. Your task is then to evaluate if the agent answer is close to the ground thruth answer. 

You are auditing the following agent:
{agent_description}

Think step by step and consider the agent output in its entirety. Remember: you need to have a strong and sound reason to support your evaluation.
If the agent answer is correct, return True. If the agent answer is incorrect, return False along with the reason.
You must output a single JSON object with keys 'correctness' and 'correctness_reason'. Make sure you return a valid JSON object.

The question that was asked to the agent, its output, and the expected ground truth answer will be delimited with XML tags.
"""

CORRECTNESS_INPUT_TEMPLATE = """<question>
{question}
</question>

<agent_answer>
{agent_answer}
</agent_answer>

<ground_truth>
{ground_truth}
</ground_truth>
"""


CORRECTNESS_TRUE_EXAMPLE_INPUT = CORRECTNESS_INPUT_TEMPLATE.format(
    question="What is the capital of France?", agent_answer="The capital of France is Paris.", ground_truth="Paris."
)

CORRECTNESS_TRUE_EXAMPLE_OUTPUT = """{"correctness": true, "correctness_reason": ""}"""

CORRECTNESS_FALSE_EXAMPLE_INPUT = CORRECTNESS_INPUT_TEMPLATE.format(
    question="What is the capital of Denmark?",
    agent_answer="The capital of Denmark is Paris.",
    ground_truth="Copenhagen.",
)

CORRECTNESS_FALSE_EXAMPLE_OUTPUT = (
    """{"correctness": false, "correctness_reason": "The capital of Denmark is Copenhagen, not Paris."}"""
)


class CorrectnessMetric(Metric):
    def __init__(self, name: str, llm_client: LLMClient = None, agent_description: Optional[str] = None):
        self.name = name
        self._llm_client = llm_client
        self.agent_description = agent_description or "This agent is a chatbot that answers question from users."

    def __call__(self, question_sample: dict, answer: AgentAnswer) -> dict:
        """
        Compute the correctness between the agent answer and the reference answer from QATestset.

        Parameters
        ----------
        question_sample : dict
            A question sample from a QATestset.
        answer : ModelOutput
            The answer of the agent on the question.

        Returns
        -------
        dict
            The result of the correctness evaluation. It contains the keys 'correctness' and 'correctness_reason'.
        """
        llm_client = self._llm_client or get_default_client()
        try:
            out = llm_client.complete(
                messages=[
                    ChatMessage(
                        role="system",
                        content=CORRECTNESS_EVALUATION_SYSTEM_PROMPT.format(agent_description=self.agent_description),
                    ),
                    ChatMessage(role="user", content=CORRECTNESS_TRUE_EXAMPLE_INPUT),
                    ChatMessage(role="assistant", content=CORRECTNESS_TRUE_EXAMPLE_OUTPUT),
                    ChatMessage(role="user", content=CORRECTNESS_FALSE_EXAMPLE_INPUT),
                    ChatMessage(role="assistant", content=CORRECTNESS_FALSE_EXAMPLE_OUTPUT),
                    ChatMessage(
                        role="user",
                        content=CORRECTNESS_INPUT_TEMPLATE.format(
                            question=question_sample.question,
                            agent_answer=answer.message,
                            ground_truth=question_sample.reference_answer,
                        ),
                    ),
                ],
                temperature=0,
                format="json",
            )
            return parse_json_output(
                out.content,
                llm_client=llm_client,
                keys=["correctness", "correctness_reason"],
                caller_id=self.__class__.__name__,
            )

        except Exception as err:
            raise LLMGenerationError("Error while evaluating the agent") from err


correctness_metric = CorrectnessMetric(name="correctness")



import tempfile
import uuid
from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from giskard import Dataset, Model, Suite
from giskard.core.suite import SuiteResult, TestPartial, TestSuiteResult, single_binary_result
from giskard.core.test_result import TestResult
from giskard.testing import test_accuracy
from tests.utils import MockedClient


@pytest.mark.parametrize(
    "results, single_result",
    [
        ([True, True, True], True),
        ([False, False, True], False),
        ([True, True, False], False),
        ([True, False, True], False),
        ([False, False, False], False),
    ],
)
def test_single_binary_result(results, single_result):
    assert single_binary_result([TestResult(passed=result) for result in results]) is single_result


def test_default_parameters_are_used_at_runtime(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    # This will miss dataset
    suite = Suite(default_params=dict(model=german_credit_model))
    suite.add_test(my_test)
    result = suite.run()
    assert not result.passed
    _, test_result, _ = result.results[0]
    assert test_result.is_error

    # But we can pass dataset at runtime
    result = suite.run(dataset=german_credit_data)
    assert result.passed
    _, test_result, _ = result.results[0]
    assert not test_result.is_error
    assert test_result.passed

    # Or we can provide it in the suite defaults
    suite.default_params["dataset"] = german_credit_data
    result = suite.run()
    assert result.passed
    _, test_result, _ = result.results[0]
    assert not test_result.is_error
    assert test_result.passed


def test_runtime_parameters_override_default_parameters(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    def constant_pred(df):
        return np.stack((np.ones(len(df)), np.zeros(len(df)))).T

    bad_model = Model(
        constant_pred, model_type="classification", classification_labels=german_credit_model.classification_labels
    )

    # The test will not pass
    suite = Suite(default_params=dict(model=bad_model, dataset=german_credit_data))
    suite.add_test(my_test)
    assert not suite.run().passed

    # With the right model, the test will pass
    assert suite.run(model=german_credit_model).passed


def test_export_for_unittest_with_default(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    suite = Suite(default_params=dict(model=german_credit_model, dataset=german_credit_data))
    suite.add_test(my_test)
    tests_list = suite.to_unittest()
    assert len(tests_list) == 1
    assert bool(tests_list[0].fullname)


def test_export_for_unittest_with_export_args(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    suite = Suite()
    suite.add_test(my_test)
    tests_list = suite.to_unittest(model=german_credit_model, dataset=german_credit_data)
    assert len(tests_list) == 1
    assert bool(tests_list[0].fullname)


def test_suite_result_backward_compatibility():
    """This allow backward compatibility by going moving Suiteresult from a tuple to a dataclass"""
    test_name, result, params = SuiteResult("name", TestResult(), {}, TestPartial(test_accuracy, dict(), 1))
    assert test_name == "name"
    assert result == TestResult()
    assert params == {}


def test_suite_result_to_dto():
    dataset = Dataset(pd.DataFrame({"test": [1, 2, 3]}))
    dataset.id = uuid.uuid4()

    suite = Suite()
    suite.id = 2

    result = TestSuiteResult(
        suite,
        {"dataset": dataset, "threshold": 0.5},
        True,
        [
            SuiteResult(
                "name", TestResult(), {"dataset": dataset, "threshold": 0.5}, TestPartial(test_accuracy, dict(), 1)
            )
        ],
        datetime.now(),
        datetime.now(),
    )

    with MockedClient() as (client, mr):
        dto = result._to_dto("label", client, "project_key")

        assert dto.inputs[0].value == str(dataset.id)
        assert dto.inputs[1].value == str(0.5)

        assert dto.results[0].inputs["dataset"] == str(dataset.id)
        assert dto.results[0].inputs["threshold"] == str(0.5)


def test_suite_save_and_load(german_credit_data, german_credit_model):
    my_test = test_accuracy(threshold=0.7)

    suite = Suite()
    suite.add_test(my_test)

    with tempfile.TemporaryDirectory() as tmp_dirname:
        suite.save(tmp_dirname)
        loaded_suite = Suite.load(tmp_dirname)

    result = loaded_suite.run(model=german_credit_model, dataset=german_credit_data)

    assert result.passed
    assert len(result.results) == 1
    _, test_result, _ = result.results[0]
    assert not test_result.is_error
    assert test_result.passed



import collections
import functools
import hashlib
import logging
import os
import sys
from pathlib import Path
from time import sleep

import click
from pydantic import AnyHttpUrl, parse_obj_as

from giskard.path_utils import run_dir

logger = logging.getLogger(__name__)
logging.getLogger("giskard").setLevel(logging.INFO)


def common_options(func):
    @click.option(
        "--verbose",
        "-v",
        is_flag=True,
        callback=set_verbose,
        default=False,
        expose_value=False,
        is_eager=True,
        help="Enable verbose logging",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def set_verbose(_ctx, _param, value):
    if value:
        logging.getLogger("giskard").setLevel(logging.DEBUG)


def remove_stale_pid_file(pid_file):
    pid = pid_file.read_pid()
    if pid is not None and is_pid_stale(pid):
        logger.debug("Stale PID file found, removing it")
        pid_file.break_lock()


def is_pid_stale(pid):
    try:
        os.kill(pid, 0)  # NOSONAR
    except (OSError, TypeError):
        return True
    else:
        return False


def create_pid_file_path(worker_name, url):
    hash_value = ml_worker_id(worker_name, url)
    return run_dir / f"ml-worker-{hash_value}.pid"


def ml_worker_id(worker_name, url):
    key = f"{sys.executable}{str(url)}{worker_name}"
    hash_value = hashlib.sha1(key.encode()).hexdigest()
    return hash_value


def validate_url(_ctx, _param, value) -> AnyHttpUrl:
    return parse_obj_as(AnyHttpUrl, value)


def get_log_path(worker_name=None):
    return run_dir / (f"ml-worker-{worker_name}.log" if worker_name else "ml-worker.log")


def tail(filename, n=100):
    """Return the last n lines of a file"""
    with Path(filename).open("r", encoding="utf-8") as fp:
        return collections.deque(fp, n)


def follow_file(filename):
    if not os.path.exists(filename):
        print(f"{filename} does not exists")
        return

    wait = 1
    with Path(filename).open("r", encoding="utf-8") as fp:
        exit_pooling = False
        skip_existing = True
        while not exit_pooling:
            line = fp.readline()

            if not line:
                skip_existing = False
                sleep(wait)
            elif not skip_existing:
                print(line, end="")



from typing import Optional

import functools
import logging
import os
import platform
import sys
from pathlib import Path

import click
import lockfile
import psutil
from click import INT, STRING
from lockfile.pidlockfile import PIDLockFile, read_pid_from_pidfile, remove_existing_pidfile
from pydantic import AnyHttpUrl

from giskard.cli_utils import (
    common_options,
    create_pid_file_path,
    follow_file,
    get_log_path,
    remove_stale_pid_file,
    tail,
    validate_url,
)
from giskard.client.giskard_client import GiskardClient
from giskard.path_utils import run_dir
from giskard.settings import settings
from giskard.utils.analytics_collector import analytics, anonymize

logger = logging.getLogger(__name__)


@click.group("worker", help="ML Worker management", context_settings={"show_default": True})
def worker() -> None:
    """
    ML Worker management
    """


def start_stop_options(func):
    @click.option(
        "--url",
        "-u",
        type=STRING,
        default="http://localhost:19000",
        help="Remote Giskard hub url",
        callback=validate_url,
    )
    @click.option(
        "--name",
        "worker_name",
        default="external_worker",
        help="Name/id of the worker starting",
    )
    @click.option(
        "--key",
        "-k",
        "api_key",
        envvar="GSK_API_KEY",
        help="Giskard hub API key",
    )
    @click.option(
        "--hf-token",
        "hf_token",
        envvar="GSK_HF_TOKEN",
        help="Access token for Giskard hosted in a private Hugging Face Spaces",
    )
    @click.option(
        "--server",
        "-s",
        "is_server",
        is_flag=True,
        default=False,
        help="Server mode. Used to control Giskard managed ML Worker",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def start_options(func):
    @click.option(
        "--parallelism",
        "nb_workers",
        default=None,
        help="Number of processes to use for parallelism (None for number of cpu)",
    )
    @click.option(
        "--daemon",
        "-d",
        "is_daemon",
        is_flag=True,
        default=False,
        help="Should ML Worker be started as a Daemon in a background",
    )
    @click.option(
        "--log-level",
        "log_level",
        default=logging.getLevelName(logging.INFO),
        type=click.Choice(
            [
                logging.getLevelName(logging.DEBUG),
                logging.getLevelName(logging.INFO),
                logging.getLevelName(logging.WARNING),
                logging.getLevelName(logging.ERROR),
            ]
        ),
        help="Global log level",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@worker.command("start")
@common_options
@start_stop_options
@start_options
def start_command(url: AnyHttpUrl, is_server, api_key, is_daemon, hf_token, nb_workers, worker_name, log_level):
    """\b
    Start ML Worker.

    ML Worker can be started in 2 modes:

    - server: Giskard starts an ML Worker managed by itself on the server.

    - client: ML Worker acts as a client and should connect to a running Giskard instance
        by specifying this instance's host and port.
    """
    analytics.track(
        "giskard-worker:start",
        {"url": anonymize(url), "log_level": log_level},
    )

    logging.root.setLevel(logging.getLevelName(log_level))

    api_key = initialize_api_key(api_key)
    hf_token = initialize_hf_token(hf_token)
    if is_server:
        # Create a Giskard client and request to start an ML worker with given name
        _start_server_command(worker_name, url, api_key, hf_token)
    else:
        _start_command(
            worker_name, url, api_key, is_daemon, hf_token, int(nb_workers) if nb_workers is not None else None
        )


def initialize_api_key(api_key):
    if not api_key:
        api_key = click.prompt("Enter Giskard hub API key", type=str)
    if "GSK_API_KEY" in os.environ:
        # delete API key environment variable so that it doesn't get leaked when the test code is executed
        del os.environ["GSK_API_KEY"]
    return api_key


def initialize_hf_token(hf_token):
    # HF token is not mandantory unless connection error
    if "GSK_HF_TOKEN" in os.environ:
        # delete HF token environment variable so that it doesn't get leaked when the test code is executed
        del os.environ["GSK_HF_TOKEN"]
    return hf_token


def _start_server_command(worker_name, url: AnyHttpUrl, api_key, hf_token=None):
    client: GiskardClient = GiskardClient(str(url), api_key, hf_token)
    client.start_managed_worker(worker_name)


def _start_command(worker_name, url: AnyHttpUrl, api_key, is_daemon, hf_token=None, nb_workers=None):
    from giskard.ml_worker.ml_worker import MLWorker

    start_msg = f"Starting ML Worker {worker_name}"
    if is_daemon:
        start_msg += " daemon"
    logger.info(start_msg)

    os.environ["TQDM_DISABLE"] = "1"
    logger.info("Python: %s (%s)", sys.executable, platform.python_version())
    logger.info("Giskard Home: %s", settings.home_dir)

    pid_file_path = create_pid_file_path(worker_name, url)
    logger.info("Creating pid file : %s", pid_file_path)

    pid_file = PIDLockFile(pid_file_path)
    remove_stale_pid_file(pid_file)

    log_path = get_log_path(worker_name=worker_name)
    logger.info(f"Writing logs to {log_path}")

    ml_worker: Optional[MLWorker] = None
    try:
        pid_file.acquire()
        if settings.force_asyncio_event_loop or sys.platform == "win32":
            logger.info("Using asyncio to run jobs")
            from asyncio import run
        else:
            logger.info("Using uvloop to run jobs")
            from uvloop import run

        if is_daemon:
            from daemon import DaemonContext
            from daemon.daemon import change_working_directory

            # Releasing the lock because it will be re-acquired by a daemon process
            pid_file.release()

            # If on windows, throw error and exit
            if sys.platform == "win32":
                logger.error("Daemon mode is not supported on Windows.")
                return

            with DaemonContext(pidfile=pid_file, stdout=open(log_path, "w+t")):
                # For some reason requests.utils.should_bypass_proxies that's called inside every request made by requests
                # hangs when the process runs as a daemon. A dirty temporary fix is to disable proxies for daemon mode.
                # True reasons for this to happen to be investigated
                os.environ["no_proxy"] = "*"

                workdir = settings.home_dir / "tmp" / f"daemon-run-{os.getpid()}"
                workdir.mkdir(exist_ok=True, parents=True)
                change_working_directory(workdir)

                logger.info(f"Daemon PID: {os.getpid()}")
                run(_start_worker(worker_name, url, api_key, hf_token, nb_workers))
        else:
            run(_start_worker(worker_name, url, api_key, hf_token, nb_workers))
    except KeyboardInterrupt:
        logger.info("Exiting")
        if ml_worker:
            ml_worker.stop()
    except lockfile.AlreadyLocked:
        existing_pid = read_pid_from_pidfile(pid_file_path)
        logger.warning(
            f"Another ML Worker {_ml_worker_description(worker_name, url)} "
            f"is already running with PID: {existing_pid}. "
            "Not starting a new one. "
            'To stop a running worker for this instance execute: "giskard worker stop" or '
            '"giskard worker stop -a" to stop all running workers'
        )
    finally:
        if pid_file.i_am_locking():
            pid_file.release()


def _ml_worker_description(worker_name: str, url: AnyHttpUrl):
    return f"named {worker_name} for {url}"


async def _start_worker(worker_name, url, api_key, hf_token, nb_workers):
    from giskard.ml_worker.ml_worker import MLWorker

    ml_worker = MLWorker(worker_name, url, api_key, hf_token)
    await ml_worker.start(restart=True, nb_workers=nb_workers)


def _stop_server_command(worker_name, url: AnyHttpUrl, api_key, hf_token=None):
    client: GiskardClient = GiskardClient(str(url), api_key, hf_token)
    return client.stop_managed_worker(worker_name)


@worker.command("stop", help="Stop running ML Workers")
@common_options
@start_stop_options
@click.option("--all", "-a", "stop_all", is_flag=True, default=False, help="Stop all running ML Workers")
def stop_command(worker_name, stop_all, is_server, url: AnyHttpUrl, api_key, hf_token=None):
    import re

    analytics.track(
        "giskard-worker:stop",
        {"url": anonymize(url), "stop_all": stop_all},
    )
    if stop_all:
        if is_server:
            logger.fatal("Stopping all workers on the server is not yet supported.")
            return
        for pid_path in run_dir.iterdir():
            if not re.match(r"^ml-worker-.*\.pid$", pid_path.name):
                continue
            _stop_pid_fname(pid_path)
    else:
        if is_server:
            api_key = initialize_api_key(api_key)
            hf_token = initialize_hf_token(hf_token)
            _stop_server_command(worker_name, url, api_key, hf_token)
        else:
            _find_and_stop(worker_name, url)


def _restart_server_command(worker_name, url: AnyHttpUrl, api_key, hf_token=None):
    client: GiskardClient = GiskardClient(url, api_key, hf_token)
    client.stop_managed_worker(worker_name)
    client.start_managed_worker(worker_name)


@worker.command("restart", help="Restart ML Worker")
@common_options
@start_stop_options
@start_options
def restart_command(url: AnyHttpUrl, is_server, api_key, is_daemon, hf_token, nb_workers, worker_name, log_level):
    analytics.track(
        "giskard-worker:restart",
        {"url": anonymize(url)},
    )
    logging.root.setLevel(logging.getLevelName(log_level))

    if is_server:
        _restart_server_command(worker_name, url, api_key, hf_token)
        return

    # Restart the local worker
    _find_and_stop(worker_name, url)
    _start_command(worker_name, url, api_key, is_daemon, hf_token=hf_token, nb_workers=nb_workers)


def _stop_pid_fname(pid_path: Path):
    pid_file_path = str(pid_path)
    remove_stale_pid_file(PIDLockFile(pid_path))
    pid = read_pid_from_pidfile(pid_path)
    if pid:
        worker_process = psutil.Process(pid)
        worker_process.terminate()
        logger.info("Stopped ML Worker by PID: %s", pid)
    else:
        logger.info("ML Worker (PID %s, path %s) is not running", pid, pid_path)
    remove_existing_pidfile(pid_file_path)


def _find_and_stop(worker_name: str, url: AnyHttpUrl):
    return _stop_pid_fname(create_pid_file_path(worker_name, url))


@worker.command("logs")
@common_options
@click.option(
    "--lines",
    "-n",
    type=INT,
    default=10,
    help="Output the last N lines of the log file, 10 lines are displayed by default",
)
@click.option(
    "--follow",
    "-f",
    "is_follow",
    is_flag=True,
    default=False,
    help="Output appended data as new logs are being generated",
)
def read_logs(lines, is_follow):
    analytics.track(
        "giskard-worker:logs",
        {
            "lines": lines,
            "is_follow": is_follow,
        },
    )
    log_path = get_log_path()

    if not os.path.exists(log_path):
        print(f"Unable to find any logfile!\n{log_path} does not exists")
        exit(-1)

    for line in tail(log_path, lines):
        print(line, end="")

    if is_follow:
        follow_file(log_path)



from typing import Dict, Optional

import os
import re
import warnings
from abc import ABC

import sentry_sdk
from requests.exceptions import ConnectionError
from sentry_sdk.integrations.excepthook import ExcepthookIntegration
from sentry_sdk.scrubber import DEFAULT_DENYLIST

WHITELISTED_MODULES = ["giskard", "pandas", "numpy"]
DISABLE_SENTRY_ENV_VARIABLE_NAME = "GSK_DISABLE_SENTRY"


def _lower_text_only_var_name(var_name: str):
    # Remove underscore and digits and put to lower case
    return re.sub(r"[0-9_]+", "", var_name.lower())


SENSITIVE_VAR_NAME = [
    _lower_text_only_var_name(var_name) for var_name in DEFAULT_DENYLIST + ["df", "batch", "requirement"]
]


def _scrub_nonsensitive_frame(frame):
    if "vars" in frame:
        frame["vars"] = {
            var_name: "[Filtered]" if _lower_text_only_var_name(var_name) in SENSITIVE_VAR_NAME else value
            for var_name, value in frame["vars"].items()
        }

    return frame


def _scrub_sensitive_frame(frame):
    if "vars" in frame:
        frame["vars"] = {var_name: "[Filtered]" for var_name in frame["vars"].keys()}

    # Anonymize code
    frame["pre_context"] = ["[filtered_pre_context]"]
    frame["context_line"] = ["[filtered_context_line]"]
    frame["post_context"] = ["[filtered_pre_context]"]

    return frame


def scrub_frame(frame):
    is_nonsensitive = (
        "module" in frame and frame["module"] is not None and frame["module"].split(".")[0] in WHITELISTED_MODULES
    )

    return _scrub_nonsensitive_frame(frame) if is_nonsensitive else _scrub_sensitive_frame(frame)


def scrub_stacktrace(stacktrace):
    if "frames" not in stacktrace:
        return stacktrace

    stacktrace["frames"] = [scrub_frame(frame) for frame in stacktrace["frames"]]

    return stacktrace


def has_giskard_module_in_frames(stacktrace):
    if "frames" not in stacktrace:
        return False

    return any(
        "module" in frame and frame["module"] is not None and frame["module"].split(".")[0] == "giskard"
        for frame in stacktrace["frames"]
    )


def scrub_event(event, _hint) -> Optional[Dict[str, ABC]]:
    if "exception" not in event:
        return event

    exception = event["exception"]
    if "values" not in exception:
        return event

    for value in exception["values"]:
        if "stacktrace" in value:
            if not has_giskard_module_in_frames(value["stacktrace"]):
                return None  # Remove noise by filtering error that arise outside of giskard (ei. in jupiter notebook)

            value["stacktrace"] = scrub_stacktrace(value["stacktrace"])

    return event


def configure_sentry():
    if os.getenv(DISABLE_SENTRY_ENV_VARIABLE_NAME, "True").lower() in ("true", "1", "yes", "t", "y"):
        return None

    try:
        sentry_sdk.init(
            # DSN is safe to be publicly available: https://docs.sentry.io/product/sentry-basics/concepts/dsn-explainer/
            dsn="https://a5d33bfa91bc3da9af2e7d32e19ff89d@o4505952637943808.ingest.sentry.io/4506789759025152",
            enable_tracing=True,
            integrations=[ExcepthookIntegration(always_run=True)],
            before_send=scrub_event,
        )
    except ConnectionError:
        warnings.warn("Failed to configure sentry due to a connection error")



"""
This module defines various test functions that can be used with pushes.

The main functions are:

- if_underconfidence_rate_decrease: Custom test for underconfidence rate decrease
- if_overconfidence_rate_decrease: Custom test for overconfidence rate decrease
- correct_example: Unit test for example correctness
- increase_probability: Test for increasing probability of ground truth label
- one_sample_overconfidence_test: One-sample test for overconfidence
- one_sample_underconfidence_test: One-sample test for underconfidence
- test_metamorphic_invariance_with_mad: Metamorphic test with mean absolute deviation

These test functions allow validating pushes and can be added to test suites.
"""
from typing import Any, Optional

from giskard.core.test_result import TestResult
from giskard.datasets.base import Dataset
from giskard.functions.transformation import mad_transformation
from giskard.models.base import BaseModel
from giskard.registry.decorators import test
from giskard.registry.slicing_function import SlicingFunction
from giskard.testing.tests.calibration import test_overconfidence_rate, test_underconfidence_rate
from giskard.testing.tests.metamorphic import test_metamorphic_invariance
from giskard.testing.tests.performance import _test_diff_prediction, test_f1, test_rmse
from giskard.testing.utils.utils import Direction, check_slice_not_empty


@test(name="Slicing RMSE difference", tags=["performance", "regression", "ground_truth", "push"])
def test_diff_rmse_push(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.1,
    direction: Direction = Direction.Decreasing,
) -> TestResult:
    """Test for difference in RMSE between a slice and full dataset.

    Checks if the RMSE on a sliced subset of the data is significantly
    different from the full dataset based on a threshold and direction.

    Can be used with pushes to test if problematic slices have worse RMSE.

    Parameters
    ----------
    model : BaseModel
        Model to test
    dataset : Dataset
        Full dataset
    slicing_function : Optional[SlicingFunction], optional
        Function to slice dataset, by default None
    threshold : float, optional
        Allowed RMSE difference, by default 0.1
    direction : Direction, optional
        Whether slice RMSE should increase or decrease, by default Direction.Decreasing

    Returns
    -------
    TestResult
        test result with pass/fail, slice sizes, RMSE diff, debug dataset
    """
    if slicing_function:
        test_name = "test_diff_rmse_push"
        actual_dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="reference_dataset", test_name=test_name)

    return _test_diff_prediction(
        test_rmse.test_fn,
        model,
        actual_dataset,
        dataset,
        threshold=threshold,
        direction=direction,
        test_name="RMSE",
    )


@test(name="Slicing f1 difference", tags=["performance", "classification", "ground_truth", "push"])
def test_diff_f1_push(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = -0.1,
    direction: Direction = Direction.Increasing,
) -> TestResult:
    """Test for difference in F1 score between a slice and full dataset.

    Checks if the F1 score on a sliced subset of the data is significantly
    different from the full dataset based on a threshold and direction.

    Can be used with pushes to test if problematic slices have worse F1 score.

    Parameters
    ----------
    model : BaseModel
        Model to test
    dataset : Dataset
        Full dataset
    slicing_function : Optional[SlicingFunction], optional
        Function to slice dataset, by default None
    threshold : float, optional
        Allowed F1 score difference, by default -0.1
    direction : Direction, optional
        Whether slice F1 should increase or decrease, by default Direction.Increasing

    Returns
    -------
    TestResult
        test result with pass/fail, slice sizes, F1 diff, debug dataset
    """
    if slicing_function:
        test_name = "test_diff_f1_push"
        actual_dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=actual_dataset, dataset_name="actual_dataset", test_name=test_name)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="reference_dataset", test_name=test_name)

    return _test_diff_prediction(
        test_f1.test_fn,
        model,
        actual_dataset,
        dataset,
        threshold=threshold,
        direction=direction,
        test_name="f1",
    )


@test(name="If Underconfidence Decreases", tags=["custom"])
def if_underconfidence_rate_decrease(model: BaseModel, dataset: Dataset, rate: float) -> TestResult:
    """Test if underconfidence rate decreases for model on dataset.

    Checks if the underconfidence rate for the model on the provided dataset
    is lower than the specified rate.

    Parameters
    ----------
    model : BaseModel
        Model to test
    dataset : Dataset
        Dataset to test on
    rate : float
        Target underconfidence rate to decrease from

    Returns
    -------
    TestResult
        Test result with pass/fail and underconfidence rate metric
    """
    new_rate = test_underconfidence_rate(model, dataset).metric
    return TestResult(passed=new_rate < rate, metric=new_rate - rate)


@test(name="If Overconfidence Decreases", tags=["custom"])
def if_overconfidence_rate_decrease(model: BaseModel, dataset: Dataset, rate: float) -> TestResult:
    """Test if overconfidence rate decreases for model on dataset.

    Checks if the overconfidence rate for the model on the provided dataset
    is lower than the specified rate.

    Parameters
    ----------
    model : BaseModel
        Model to test
    dataset : Dataset
        Dataset to test on
    rate : float
        Target overconfidence rate to decrease from

    Returns
    -------
    TestResult
        Test result with pass/fail and overconfidence rate metric
    """
    new_rate = test_overconfidence_rate(model, dataset).metric
    return TestResult(passed=new_rate < rate, metric=new_rate - rate)


@test(name="Example Correctness", tags=["unit test", "custom"])
def correct_example(model: BaseModel, saved_example: Dataset, training_label: Any) -> TestResult:
    """Test if model correctly predicts example.

    Checks if the model's prediction on the saved example matches the
    provided training label ground truth.

    Parameters
    ----------
    model : BaseModel
        Model to test
    saved_example : Dataset
        Example dataset
    training_label : Any
        Ground truth label

    Returns
    -------
    TestResult
        Test result with pass/fail
    """
    prediction = model.predict(saved_example).prediction.values[0]
    return TestResult(passed=prediction == training_label, metric=prediction == training_label)


@test(name="Increase Probability", tags=["unit test", "custom"])
def increase_probability(
    model: BaseModel, saved_example: Dataset, training_label: Any, training_label_proba: Any
) -> TestResult:
    """Test if model probability for ground truth label increases.

    Checks if the predicted probability for the training label increases
    compared to the original training probability.

    Parameters
    ----------
    model : BaseModel
        Model to test
    saved_example : Dataset
        Example dataset
    training_label : Any
        Ground truth label
    training_label_proba : Any
        Original probability

    Returns
    -------
    TestResult
        Test result with pass/fail and probability metric
    """
    proba = model.predict(saved_example).all_predictions[training_label].values[0]
    return TestResult(passed=proba > training_label_proba, metric=proba - training_label_proba)


@test(name="One-Sample Overconfidence test", tags=["one-sample test", "custom"])
def one_sample_overconfidence_test(model: BaseModel, saved_example: Dataset) -> TestResult:
    """One-sample overconfidence test for example.

    Checks if the overconfidence rate for the model on the
    provided example is below a threshold.

    Parameters
    ----------
    model : BaseModel
        Model to test
    saved_example : Dataset
        Example dataset

    Returns
    -------
    TestResult
        Test result with pass/fail and overconfidence rate metric
    """
    if model.is_classification:
        test_result = test_overconfidence_rate(model, saved_example).execute()
        return TestResult(passed=test_result.passed, metric=test_result.metric)


@test(name="One-Sample Underconfidence test", tags=["one-sample test", "custom"])
def one_sample_underconfidence_test(model: BaseModel, saved_example: Dataset) -> TestResult:
    """One-sample underconfidence test for example.

    Checks if the underconfidence rate for the model on the
    provided example is below a threshold.

    Parameters
    ----------
    model : BaseModel
        Model to test
    saved_example : Dataset
        Example dataset

    Returns
    -------
    TestResult
        Test result with pass/fail and underconfidence rate metric
    """
    if model.is_classification:
        test_result = test_underconfidence_rate(model, saved_example).execute()
        return TestResult(passed=test_result.passed, metric=test_result.metric)


@test(name="Numerical Invariance test", tags=["custom"])
def test_metamorphic_invariance_with_mad(
    model: BaseModel, dataset: Dataset, column_name: str, value_added: float
) -> TestResult:
    """Metamorphic test for numerical invariance with mean absolute deviation.

    Checks if adding a value to a numerical feature keeps the prediction
    approximately the same using mean absolute deviation.

    Parameters
    ----------
    model : BaseModel
        Model to test
    dataset : Dataset
        Dataset to test
    column_name : str
        Name of numerical feature column
    value_added : float
        Value to add to column

    Returns
    -------
    TestResult
        Test result with pass/fail
    """
    return test_metamorphic_invariance(
        model=model,
        dataset=dataset,
        transformation_function=mad_transformation(column_name=column_name, value_added=value_added),
    ).execute()
